#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <vector>
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <boost/scoped_array.hpp>
#include <boost/foreach.hpp>
#include <boost/program_options.hpp>
#include <boost/timer.hpp>
#include "entropy.h"
#include "pe32.h"
#include "unpacker.h"

#define STATUS_SUCCESS 0x00000000L
#define STATUS_BUFFER_TOO_SMALL 0xC0000023L
#define STATUS_BUFFER_ALL_ZEROS 0x00000117L

//Prototypes of functions that are used to compress data
typedef DWORD (__stdcall *RtlCompressBuffer_Fn)(
	unsigned short CompressionFormatAndEngine,
	char* UncompressedBuffer,
	unsigned long UncompressedBufferSize,
	char* CompressedBuffer,
	unsigned long CompressedBufferSize,
	unsigned long UncompressedChunkSize,
	unsigned long* FinalCompressedSize,
	void* WorkSpace);

typedef DWORD (__stdcall *RtlGetCompressionWorkSpaceSize_Fn)(
	unsigned short CompressionFormatAndEngine,
	unsigned long* CompressBufferWorkSpaceSize,
	unsigned long* CompressFragmentWorkSpaceSize);

static RtlCompressBuffer_Fn RtlCompressBuffer_;
static RtlGetCompressionWorkSpaceSize_Fn RtlGetCompressionWorkSpaceSize_;

//Macros to align TLS data
#define TLS_ALIGNMENT (2 * sizeof(void *))
#define TLS_ALIGN(size) (((size) + TLS_ALIGNMENT - 1) & ~(TLS_ALIGNMENT - 1))

//Structures representing icon group and icon header inside PE resources
#pragma pack(push, 1)
struct ICON_GROUP
{
	BYTE Width;
	BYTE Height;
	BYTE ColorCount;
	BYTE Reserved;
	WORD Planes;
	WORD BitCount;
	DWORD SizeInBytes;
	WORD Number; //Represents resource ID in PE icon list
};
#pragma pack(pop, 1)

#pragma pack(push, 1)
struct ICO_HEADER
{
	WORD Reserved;
	WORD Type;
	WORD Count; //Count of icons included in icon group
};
#pragma pack(pop, 1)

//Unpacker section name, used to determine if PE is already packed
static const std::string unpacker_section_name = "kaimi.ru";

//Structure representing PE exported function
struct exported_function
{
	exported_function()
		:ordinal(0), rva(0), has_name(false), name_ordinal(0), redirect(false)
	{}

	WORD ordinal; //Function ordinal
	DWORD rva; //Function RVA
	std::string name; //Function name
	bool has_name; //true == function has name
	WORD name_ordinal; //Function name ordinal
	bool redirect; //true == function is forwarded
	std::string redirect_name; //Name of forwarded function
};

//Operator used to compare to functions to sort them alphabetically
bool operator<(const exported_function& func1, const exported_function& func2)
{
	if(func1.has_name && func2.has_name)
		return func1.name < func2.name;
	else
		return func1.ordinal < func2.ordinal;
}

//Function to get specified PE resource, returns true if resource was found
//This function has no checks for resource loops (they're hardly possible and can't be in valid PE files)
bool get_resource(const pe32& image, //PE image
	DWORD res_type, //Type ID of resource
	IMAGE_RESOURCE_DATA_ENTRY& ret, //Data entry of resource
	IMAGE_RESOURCE_DIRECTORY_ENTRY& type_dir, //Type directory entry
	IMAGE_RESOURCE_DIRECTORY_ENTRY& name_dir, //Name directory entry
	IMAGE_RESOURCE_DIRECTORY& name_res, //Name directory
	IMAGE_RESOURCE_DIRECTORY_ENTRY& lang_dir, //Language directory entry
	IMAGE_RESOURCE_DIRECTORY& lang_res, //Language directory
	unsigned long name_number = 0, //Needed number of name directory entry
	unsigned long lang_number = 0) //Needed number of language directory entry
{
	DWORD res_rva = image.directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE);

	if(image.section_data_length_from_rva(res_rva) < sizeof(IMAGE_RESOURCE_DIRECTORY))
		throw pe_exception("Incorrect resource directory");

	//Root resource directory
	const IMAGE_RESOURCE_DIRECTORY* resources = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY*>(
		image.section_data_from_rva(res_rva));

	unsigned long total_resources = resources->NumberOfIdEntries + resources->NumberOfNamedEntries;

	for(unsigned long i = 0; i < total_resources; i++)
	{
		//"Type" level
		if(image.section_data_length_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + i * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)) < sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))
			throw pe_exception("Incorrect resource directory");

		const IMAGE_RESOURCE_DIRECTORY_ENTRY* entry_type = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY_ENTRY*>(
			image.section_data_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + i * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)));

		//Check resource type
		if(!entry_type->NameIsString && entry_type->Name == res_type)
		{
			type_dir = *entry_type;

			if(!entry_type->DataIsDirectory)
				throw pe_exception("Incorrect resource directory");

			//"Name" level
			{
				if(image.section_data_length_from_rva(res_rva + entry_type->OffsetToDirectory) < sizeof(IMAGE_RESOURCE_DIRECTORY))
					throw pe_exception("Incorrect resource directory");

				const IMAGE_RESOURCE_DIRECTORY* nameres = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY*>(
					image.section_data_from_rva(res_rva + entry_type->OffsetToDirectory));

				name_res = *nameres;
			}

			if(name_number >= static_cast<unsigned long>(name_res.NumberOfIdEntries) + static_cast<unsigned long>(name_res.NumberOfNamedEntries))
				return false;

			if(image.section_data_length_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + entry_type->OffsetToDirectory + name_number * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)) < sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))
				throw pe_exception("Incorrect resource directory");

			const IMAGE_RESOURCE_DIRECTORY_ENTRY* entry_name = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY_ENTRY*>(
				image.section_data_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + entry_type->OffsetToDirectory + name_number * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)));

			name_dir = *entry_name;

			if(!entry_name->OffsetToData)
				return false;

			if(!entry_name->DataIsDirectory)
				throw pe_exception("Incorrect resource directory");

			//"Language" level
			{
				if(image.section_data_length_from_rva(res_rva + entry_name->OffsetToDirectory) < sizeof(IMAGE_RESOURCE_DIRECTORY))
					throw pe_exception("Incorrect resource directory");

				const IMAGE_RESOURCE_DIRECTORY* langres = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY*>(
					image.section_data_from_rva(res_rva + entry_name->OffsetToDirectory));

				lang_res = *langres;
			}
			
			if(lang_number >= static_cast<unsigned long>(lang_res.NumberOfIdEntries) + static_cast<unsigned long>(lang_res.NumberOfNamedEntries))
				return false;

			if(image.section_data_length_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + entry_name->OffsetToDirectory + lang_number * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)) < sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))
				throw pe_exception("Incorrect resource directory");

			const IMAGE_RESOURCE_DIRECTORY_ENTRY* entry_language = reinterpret_cast<const IMAGE_RESOURCE_DIRECTORY_ENTRY*>(
				image.section_data_from_rva(res_rva + sizeof(IMAGE_RESOURCE_DIRECTORY) + entry_name->OffsetToDirectory + lang_number * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)));

			if(!entry_name->OffsetToData)
				return false;

			lang_dir = *entry_language;

			if(!entry_language->DataIsDirectory)
			{
				//Store resource data entry
				if(image.section_data_length_from_rva(res_rva + entry_language->OffsetToData) < sizeof(IMAGE_RESOURCE_DATA_ENTRY))
					throw pe_exception("Incorrect resource directory");

				ret = *reinterpret_cast<const IMAGE_RESOURCE_DATA_ENTRY*>(image.section_data_from_rva(res_rva + entry_language->OffsetToData));
				return true;
			}
			else
			{
				throw pe_exception("Incorrect resource directory");
			}
		}
	}

	return false;
}

int main(int argc, char* argv[])
{
	//Parse command line
	namespace po = boost::program_options;
	
	boost::timer tmr;
	
	bool pack_res;
	bool force = false;
	bool strip_dos_headers;
	unsigned long file_alignment;
	std::string input_file_name;
	std::vector<std::string> pack_sections;
	bool add_import = false;
	bool move_icons;
	
	std::string output_file_name;
	
	po::options_description visible_options("Simple Packer by DX");

	try
	{
		po::options_description cmdline;

		visible_options.add_options() 
			("out-file,o", po::value<std::string>(&output_file_name), "Output file name")
			("add-import,i", "Add fake import table to packed image")
			("pack-rsrc,r", po::value<bool>(&pack_res)->default_value(true), "Pack .rsrc section (image resources)")
			("move-icons,m", po::value<bool>(&move_icons)->default_value(true), "Move application icons to new .rsrc")
			("pack-sections,s", po::value<std::vector<std::string>>(&pack_sections)->multitoken(), "Explicit sections name(s) to pack")
			("file-align,a", po::value<unsigned long>(&file_alignment)->default_value(512), "Packed file alignment")
			("strip-dos,d", po::value<bool>(&strip_dos_headers)->default_value(false), "Strip DOS headers")
			("force,f", "Force packing of possibly incorrect binaries")
			;

		cmdline.add(visible_options);
		cmdline.add_options()
			("image", po::value<std::string>(&input_file_name), "PE image to pack")
			;    

		po::positional_options_description desc_pos;
		desc_pos.add("image", 1);

		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).
			options(cmdline).positional(desc_pos).run(), vm);
		po::notify(vm);

		if(input_file_name.empty())
			throw std::exception("No input file specified");
		
		if(vm.count("force"))
			force = true;

		if(vm.count("add-import"))
			add_import = true;
	}
	catch(const std::exception& e)
	{
		std::cout << e.what() << std::endl << std::endl;
		std::cout << visible_options << std::endl;
		system("pause");
		return 0;
	}


	//Prepare compression functions from ntdll.dll
	HMODULE ntdll_lib = LoadLibrary(L"ntdll.dll");
	if(ntdll_lib != NULL)
	{
		RtlCompressBuffer_ = reinterpret_cast<RtlCompressBuffer_Fn>(GetProcAddress(ntdll_lib, "RtlCompressBuffer"));
		RtlGetCompressionWorkSpaceSize_ = reinterpret_cast<RtlGetCompressionWorkSpaceSize_Fn>(GetProcAddress(ntdll_lib, "RtlGetCompressionWorkSpaceSize"));

		if(!RtlCompressBuffer_ || !RtlGetCompressionWorkSpaceSize_)
		{
			std::cout << "Can't locate compression functions inside ntdll.dll" << std::endl;
			system("pause");
			return 0;
		}
	}
	else
	{
		std::cout << "Can't load ntdll.dll" << std::endl;
		return 0;
	}

	//Open PE file
	std::cout << "Packing " << input_file_name << "..." << std::endl;
	std::ifstream pe_file;
	pe_file.open(input_file_name.c_str(), std::ios::in | std::ios::binary);
	if(!pe_file)
	{
		std::cout << "Can't open PE file" << std::endl;
		system("pause");
		return 0;
	}

	//Calculate entropy to check if file was already compressed
	std::cout << "Calculating entropy...";
	double compr = entropy::possible_compression(pe_file);
	std::cout << " (" << compr << ")" << std::endl;
	pe_file.clear();

	pe_file.seekg(0, std::ios::end);
	unsigned long pe_file_size = static_cast<unsigned long>(pe_file.tellg());
	pe_file.seekg(0, std::ios::beg);
	pe_file.clear();

	try
	{
		if(!force && compr > 0.95)
			throw pe_exception("The file has been probably already compressed!");

		//Load image into memory
		std::cout << "Loading PE image into memory..." << std::endl;
		pe32 image(pe_file);
		pe_file.close();
		
		//Don't allow to pack drivers etc
		if(!force && image.get_subsystem() != IMAGE_SUBSYSTEM_WINDOWS_GUI && image.get_subsystem() != IMAGE_SUBSYSTEM_WINDOWS_CUI)
			throw pe_exception("Unsupported image subsystem");

		//Don't allow to pack .NET executables
		if(!force && image.is_dotnet())
			throw pe_exception("Image is .NET. Can't pack it.");

		{
			std::vector<std::string> section_names;

			//Check if PE file has been already packed by us
			for(std::vector<pe32::section>::const_iterator it = image.get_image_sections().begin(); it != image.get_image_sections().end(); ++it)
			{
				if(!force && (*it).get_name() == unpacker_section_name)
					throw pe_exception("File was already packed!");

				section_names.push_back((*it).get_name());
			}

			if(!pack_sections.empty())
			{
				//Check if each section from command line exists
				for(std::vector<std::string>::const_iterator it = pack_sections.begin(); it != pack_sections.end(); it++)
				{
					if(std::find(section_names.begin(), section_names.end(), *it) == section_names.end())
						throw pe_exception("Section \"" + (*it) + "\" doesn't exist!");
				}
			}
		}

		if(image.get_image_sections().back().header.PointerToRawData + image.get_image_sections().back().get_raw_data().size() < pe_file_size)
			std::cout << "WARNING!" << std::endl << "This image probably has overlay and can become unworkable after packing!" << std::endl << std::endl;
		
		//Realign executable
		std::cout << "Realigning executable..." << std::endl;
		image.set_file_alignment(file_alignment);

		//Strip VC++ stub overlay, if any
		std::cout << "Stripping VC++ stub overlay (if exists)..." << std::endl;
		image.strip_stub_overlay();

		//Pointer to added to PE file unpacker section (will be used later)
		pe32::section* added_section;

		std::size_t size_of_additional_data = 0, //size of TLS, fixups, exports in total
			size_of_tls = 0,
			size_of_fixups = 0,
			size_of_exports = 0;

		std::size_t size_of_export_relocs = 0;

		//Unpacker section data length after stripping nullbytes
		std::string::size_type actual_data_length;

		//Used if we're adding fake import
		unsigned long import_place = 0, old_import_place = 0;

		{
			//Create unpacker section
			pe32::section unpack_section;
			unpack_section.set_name(unpacker_section_name);

			//Turn on all attributes
			unpack_section.writeable(true).executable(true).readable(true);

			//Binary data to use in unpacker section
			std::string& unpacker_raw_data = unpack_section.get_raw_data();

			std::cout << std::showbase << std::hex;

			//If our image has static TLS, we should save it
			if(image.has_tls())
			{
				if(image.section_data_length_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_TLS)) < sizeof(IMAGE_TLS_DIRECTORY))
					throw pe_exception("Incorrect TLS directory");

				const IMAGE_TLS_DIRECTORY* tls = reinterpret_cast<const IMAGE_TLS_DIRECTORY*>
					(image.section_data_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_TLS)));

				std::cout << "Image has TLS on RVA " << image.directory_rva(IMAGE_DIRECTORY_ENTRY_TLS) << std::endl;

				//Copy TLS directory to unpacker section
				unpacker_raw_data.append(reinterpret_cast<const char*>(tls), sizeof(IMAGE_TLS_DIRECTORY));

				//Add empty DWORD to store TLS index
				unpacker_raw_data.append(sizeof(DWORD), '\0');

				//TLS can have raw size less than actual size of section containing it
				//So we will copy TLS data to the end of unpacker section later
				//If TLS has SizeOfZeroFill != 0, it will be OK too

				//Store size of TLS data
				size_of_tls = unpacker_raw_data.length();
				size_of_additional_data += size_of_tls;
			}

			//Original fixup table address
			DWORD original_fixup_table = 0;

			//If image has fixups
			if(image.has_reloc())
			{
				//Create empty relocation section
				//We'll set fixup table RVA later
				original_fixup_table = image.directory_rva(IMAGE_DIRECTORY_ENTRY_BASERELOC);

				std::cout << "Image has fixups on RVA " << original_fixup_table << std::endl;

				//If image has TLS, we should add fixups for it, too
				if(image.has_tls())
				{
					IMAGE_BASE_RELOCATION reloc = {0};

					reloc.SizeOfBlock = sizeof(IMAGE_BASE_RELOCATION) + 3 * sizeof(WORD); //3 fixups for AddressOfIndex, EndAddressOfRawData, StartAddressOfRawData
					//reloc.VirtualAddress will be filled in lated
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc), sizeof(reloc));

					WORD reloc_elem = 0x3000; //relocation relative (type 3)
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc_elem), sizeof(WORD));
					reloc_elem += sizeof(DWORD);
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc_elem), sizeof(WORD));
					reloc_elem += sizeof(DWORD);
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc_elem), sizeof(WORD));

					size_of_fixups = reloc.SizeOfBlock;
				}

				{
					IMAGE_BASE_RELOCATION reloc = {0};

					//Add a relocation entry to make loader fix our fixed_image_base
					//to know actual image base VA from unpacker
					reloc.SizeOfBlock = sizeof(IMAGE_BASE_RELOCATION) + sizeof(WORD);
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc), sizeof(reloc));
					WORD reloc_elem = 0x3000;
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc_elem), sizeof(WORD));
					size_of_fixups += reloc.SizeOfBlock;
				}


				{
					IMAGE_BASE_RELOCATION reloc = {0};

					//Add an empty relocation entry
					reloc.SizeOfBlock = sizeof(IMAGE_BASE_RELOCATION);
					unpacker_raw_data.append(reinterpret_cast<const char*>(&reloc), sizeof(reloc));
					size_of_fixups += sizeof(IMAGE_BASE_RELOCATION);
				}

				size_of_additional_data += size_of_fixups;
				image.set_directory_size(IMAGE_DIRECTORY_ENTRY_BASERELOC, size_of_fixups);
			}


			//Ordinal, exported function
			std::map<WORD, exported_function> functions;
			//Place for export table
			unsigned long export_place = 0,
				library_name_length = 0;
			if(image.has_exports())
			{
				if(image.section_data_length_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT)) < sizeof(IMAGE_EXPORT_DIRECTORY))
					throw pe_exception("Incorrect export directory");

				const IMAGE_EXPORT_DIRECTORY* exports = reinterpret_cast<const IMAGE_EXPORT_DIRECTORY*>
					(image.section_data_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT)));

				std::cout << "Image has exports on RVA " << image.directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT) << std::endl
					<< "Number of exported functions: " << exports->NumberOfFunctions << std::endl
					<< "Reading exports..." << std::endl;
				
				if(!exports->NumberOfFunctions || exports->NumberOfNames > exports->NumberOfFunctions)
					throw pe_exception("Incorrect export directory");

				if((!exports->AddressOfNameOrdinals && exports->AddressOfNames) ||
					(exports->AddressOfNameOrdinals && !exports->AddressOfNames))
					throw pe_exception("Incorrect export directory");

				if(image.section_data_length_from_rva(exports->AddressOfFunctions) < exports->NumberOfFunctions * sizeof(DWORD))
					throw pe_exception("Incorrect export directory");

				if(exports->AddressOfNames)
				{
					if(image.section_data_length_from_rva(exports->AddressOfNameOrdinals) < exports->NumberOfNames * sizeof(WORD))
						throw pe_exception("Incorrect export directory");

					if(image.section_data_length_from_rva(exports->AddressOfNames) < exports->NumberOfNames * sizeof(DWORD))
						throw pe_exception("Incorrect export directory");
				}

				for(std::size_t ordinal = 0; ordinal < exports->NumberOfFunctions; ordinal++)
				{
					//Get function address
					DWORD address = *reinterpret_cast<DWORD*>(
						image.section_data_from_rva(exports->AddressOfFunctions + ordinal * sizeof(DWORD)));

					//If we have a skip
					if(!address)
						continue;

					exported_function func;
					func.rva = address;

					//Scan for function name ordinal
					for(std::size_t i = 0; i < exports->NumberOfNames; i++)
					{
						WORD ordinal2 = *reinterpret_cast<WORD*>(
							image.section_data_from_rva(exports->AddressOfNameOrdinals + i * sizeof(WORD)));

						//If function has name (and name ordinal)
						if(ordinal == ordinal2)
						{
							//Get function name
							std::string::size_type max_name_length;
							if((max_name_length = image.section_data_length_from_rva(*reinterpret_cast<DWORD*>(
								image.section_data_from_rva(exports->AddressOfNames + i * sizeof(DWORD))
								))) < 1)
								throw pe_exception("Incorrect export directory");

							//Get function name
							const char* func_name = 
								image.section_data_from_rva(*reinterpret_cast<DWORD*>(
								image.section_data_from_rva(exports->AddressOfNames + i * sizeof(DWORD))
								));

							//Check for null-termination
							bool null_byte_found = 0;
							for(std::string::size_type l = 0; l < max_name_length; l++)
							{
								if(func_name[l] == '\0')
								{
									null_byte_found = true;
									break;
								}
							}

							if(!null_byte_found)
								throw pe_exception("Incorrect export directory");

							//Save function info
							func.name = func_name;
							func.name_ordinal = ordinal2;
							func.has_name = true;

							//If the function is just a redirect, save its name
							if(address >= image.directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT) + sizeof(IMAGE_DIRECTORY_ENTRY_EXPORT) &&
								address < image.directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT) + image.directory_size(IMAGE_DIRECTORY_ENTRY_EXPORT))
							{
								func.redirect_name = image.section_data_from_rva(address);
								func.redirect = true;
							}

							break;
						}
					}

					//Some duplication, we'll need it later to sort functions
					func.ordinal = ordinal;
					functions.insert(std::make_pair(ordinal, func));
				}

				//Copy export directory to unpacker section
				unpacker_raw_data.append(reinterpret_cast<const char*>(exports), sizeof(IMAGE_EXPORT_DIRECTORY));

				//Copy name of library
				{
					std::string libname = image.section_data_from_rva(exports->Name);
					library_name_length = libname.length() + 1;
					unpacker_raw_data.append(libname.c_str(), library_name_length);
					size_of_exports = sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length;
				}

				//Reserve place for export table
				unsigned long ordinal = 0;
				for(std::map<WORD, exported_function>::const_iterator it = functions.begin(); it != functions.end(); ++it)
				{
					//Don't forget to add ordinal spacing
					if((*it).first != ordinal)
					{
						for(unsigned long j = ordinal; j < (*it).first; j++)
						{
							export_place += sizeof(DWORD);
							ordinal++;
						}
					}

					export_place += sizeof(DWORD); //rva
					ordinal++;

					if((*it).second.has_name)
					{
						export_place += (*it).second.name.length() + 1;
						export_place += sizeof(WORD); //name ordinal
						export_place += sizeof(DWORD); //name address

						if((*it).second.redirect)
							export_place += (*it).second.redirect_name.length() + 1;
					}
				}

				unpacker_raw_data.append(export_place, '\0');
				size_of_exports += export_place;
				size_of_additional_data += size_of_exports;
			}

			unpacker_raw_data.append(reinterpret_cast<const char*>(&unpackdata[0]), sizeof(unpackdata));

			//Strip ending unpacker nullbytes
			for(actual_data_length = unpacker_raw_data.length() - 1; actual_data_length > 0; actual_data_length--)
			{
				if(unpacker_raw_data[actual_data_length - 1] != '\0')
					break;
			}


			//If we're asked to add import table, reserve some place for it
			if(add_import)
			{
				import_place = actual_data_length;
				actual_data_length += sizeof(IMAGE_IMPORT_DESCRIPTOR) * 2 /* One for empty element */;
				actual_data_length += 13; //"kernel32.dll" name and nullbyte
				actual_data_length += sizeof(DWORD) * 2; //Import Lookup Table + 1 empty element
				actual_data_length += sizeof(WORD); //from IMAGE_IMPORT_BY_NAME
				actual_data_length += 9; //"lstrlenA" name
				actual_data_length += sizeof(DWORD) * 2; //Import Address Table + 1 empty element
				actual_data_length += sizeof(DWORD); //DWORD where GetDC address will be written
			}
			
			unpacker_raw_data.resize(actual_data_length, '\0');


			//Save address of original fixup table (if present)
			*reinterpret_cast<DWORD*>(&unpacker_raw_data[fixup_table + size_of_additional_data]) = original_fixup_table;



			//Add TLS data to the end of section
			if(image.has_tls())
			{
				const IMAGE_TLS_DIRECTORY* tls = reinterpret_cast<const IMAGE_TLS_DIRECTORY*>
					(image.section_data_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_TLS)));

				unsigned long copy_size = std::min<unsigned long>(image.section_data_length_from_va(tls->StartAddressOfRawData), tls->EndAddressOfRawData - tls->StartAddressOfRawData);
				if(copy_size)
					unpacker_raw_data.append(image.section_data_from_va(tls->StartAddressOfRawData), copy_size);
			}


			//Repack resources (manifest, version info, icons)
			//And add them to a new .rsrc section
			if(pack_res && image.has_resources())
			{
				std::cout << "Image has resources on RVA " << image.directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE) << std::endl;
				std::cout << "Rebuilding resources..." << std::endl;

				std::string rsrc_section_data;

				//Determine, how much space we need to store "Type" tables
				DWORD total_type_space = 0;
				//Count of needed type resource directories
				unsigned long type_count = 0,
					//Icon count (if image has icons)
					icon_count = 0;

				{
					IMAGE_RESOURCE_DATA_ENTRY data;
					IMAGE_RESOURCE_DIRECTORY_ENTRY t, n, l;
					IMAGE_RESOURCE_DIRECTORY nr, lr;
					//Get first icon group resource
					//Windows uses it to determine which icons should be shown as application icons
					if(move_icons && get_resource(image, reinterpret_cast<DWORD>(RT_GROUP_ICON), data, t, n, nr, l, lr))
					{
						type_count += 2; //Two type directories for RT_ICON and RT_GROUP_ICON
						
						if(image.section_data_length_from_rva(data.OffsetToData) < sizeof(ICO_HEADER))
							throw pe_exception("Incorrect recource directory");

						//Get application icon count
						icon_count = reinterpret_cast<const ICO_HEADER*>(image.section_data_from_rva(data.OffsetToData))->Count;
						
						if(!icon_count || image.section_data_length_from_rva(data.OffsetToData + sizeof(ICO_HEADER)) < sizeof(ICON_GROUP) * icon_count)
							throw pe_exception("Incorrect recource directory");
						
						total_type_space += data.Size;
						
						const ICON_GROUP* group = reinterpret_cast<const ICON_GROUP*>(image.section_data_from_rva(data.OffsetToData + sizeof(ICO_HEADER)));
						//Enumerate icons
						for(unsigned long i = 0; i < icon_count; i++)
						{
							unsigned long cur = 0;
							
							//Search for icon with needed ID
							while(true)
							{
								if(!get_resource(image, reinterpret_cast<DWORD>(RT_ICON), data, t, n, nr, l, lr, cur))
									throw pe_exception("Incorrect resource directory");

								if(n.Id == group->Number)
									break;

								cur++;
							}

							group++;
							
							//Calculate total sizeo of needed icons
							total_type_space += data.Size;
						}

						group -= icon_count;
					}

					//If image has version info
					if(get_resource(image, reinterpret_cast<DWORD>(RT_VERSION), data, t, n, nr, l, lr))
					{
						type_count++;
						total_type_space += data.Size;
					}

					//If image has manifest
					if(get_resource(image, reinterpret_cast<DWORD>(RT_MANIFEST), data, t, n, nr, l, lr))
					{
						type_count++;
						total_type_space += data.Size;
					}
				}

				//Rename old .rsrc section
				for(std::vector<pe32::section>::iterator it = image.get_image_sections().begin(); it != image.get_image_sections().end(); ++it)
				{
					if((*it).get_name() == ".rsrc")
					{
						(*it).set_name("oldres");
						break;
					}
				}

				//Set up unpacker resources addresses to later restore original resources in image header
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[res_data_dir_addr + size_of_additional_data]) = 0xC8; //always after packing
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[original_res_addr + size_of_additional_data]) = image.directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE);
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[original_res_size + size_of_additional_data]) = image.directory_size(IMAGE_DIRECTORY_ENTRY_RESOURCE);


				//If we have some resources to repack
				if(total_type_space)
				{
					//Determine offsets for each resource level
					DWORD type_level_offset = sizeof(IMAGE_RESOURCE_DIRECTORY);
					DWORD name_level_offset = type_level_offset + type_count * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY) + sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
					DWORD lang_level_offset = name_level_offset + type_count * (sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY) * 2 + sizeof(IMAGE_RESOURCE_DIRECTORY));

					if(icon_count)
						lang_level_offset += (icon_count - 1) * sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);

					DWORD data_offset = lang_level_offset + type_count * (sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY) * 2 + sizeof(IMAGE_RESOURCE_DIRECTORY));

					if(icon_count)
						data_offset += (icon_count - 1) * (sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY) * 2 + sizeof(IMAGE_RESOURCE_DIRECTORY));

					DWORD res_offset = data_offset + type_count * sizeof(IMAGE_RESOURCE_DATA_ENTRY);

					if(icon_count)
						res_offset += (icon_count - 1) * sizeof(IMAGE_RESOURCE_DATA_ENTRY);

					total_type_space += res_offset;
					rsrc_section_data.resize(total_type_space, '\0');


					//Add new .rsrc section
					pe32::section* rsrc_section;

					{
						pe32::section s;
						s.set_name(".rsrc");
						s.readable(true).writeable(false).executable(false);
						s.get_raw_data() = rsrc_section_data;
						rsrc_section = &image.add_section(s);
					}

					std::string& rsrc_data = rsrc_section->get_raw_data();

					//Copy directory to new section
					memcpy(&rsrc_data[0], image.section_data_from_rva(image.directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE)), sizeof(IMAGE_RESOURCE_DIRECTORY));

					//Set up root type resource directory
					{
						PIMAGE_RESOURCE_DIRECTORY res = reinterpret_cast<PIMAGE_RESOURCE_DIRECTORY>(&rsrc_data[0]);
						res->NumberOfIdEntries = static_cast<WORD>(type_count);
						res->NumberOfNamedEntries = 0;
					}
					
					
					//Repack resources
					IMAGE_RESOURCE_DIRECTORY name_res, lang_res;
					IMAGE_RESOURCE_DATA_ENTRY data;
					IMAGE_RESOURCE_DIRECTORY_ENTRY type_entry, name_entry, lang_entry;

					//If image has icon(s)
					if(icon_count)
					{
						get_resource(image, reinterpret_cast<DWORD>(RT_GROUP_ICON), data, type_entry, name_entry, name_res, lang_entry, lang_res);
						const ICON_GROUP* group = reinterpret_cast<const ICON_GROUP*>(image.section_data_from_rva(data.OffsetToData + sizeof(ICO_HEADER)));

						//Enumerate icons and store them to our resorce section
						for(unsigned long i = 0; i < icon_count; i++)
						{
							unsigned long cur = 0;

							//Search icon by needed ID
							while(true)
							{
								if(!get_resource(image, reinterpret_cast<DWORD>(RT_ICON), data, type_entry, name_entry, name_res, lang_entry, lang_res, cur))
									throw pe_exception("Incorrect Resource directory");

								if(name_entry.Id == group->Number)
									break;

								cur++;
							}

							group++;

							//Set up directory entries offsets
							name_entry.OffsetToData = 0x80000000 | lang_level_offset; //To directory
							type_entry.OffsetToData = 0x80000000 | name_level_offset; //To directory
							lang_entry.OffsetToData = data_offset; //To data
							
							//Set up resource language directory and copy it to our section for each icon
							lang_res.NumberOfIdEntries = 1;
							lang_res.NumberOfNamedEntries = 0;
							memcpy(&rsrc_data[lang_level_offset], &lang_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);

							//Add one name resource directory
							if(i == 0)
							{
								memcpy(&rsrc_data[type_level_offset], &type_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
								type_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);

								name_res.NumberOfIdEntries = static_cast<WORD>(icon_count);
								name_res.NumberOfNamedEntries = 0;
								memcpy(&rsrc_data[name_level_offset], &name_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
								name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);
							}

							//Set up name as ID
							name_entry.Name = i + 1;
							memcpy(&rsrc_data[name_level_offset], &name_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
							memcpy(&rsrc_data[lang_level_offset], &lang_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
							memcpy(&rsrc_data[res_offset], image.section_data_from_rva(data.OffsetToData), data.Size);

							//Copy data entry
							data.OffsetToData = rsrc_section->header.VirtualAddress + res_offset;
							memcpy(&rsrc_data[data_offset], &data, sizeof(IMAGE_RESOURCE_DATA_ENTRY));

							name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
							data_offset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);
							res_offset += data.Size;

							//If last icon
							if(i == icon_count - 1)
								name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element

							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element
						}

						//Now add icon group to our resource section
						get_resource(image, reinterpret_cast<DWORD>(RT_GROUP_ICON), data, type_entry, name_entry, name_res, lang_entry, lang_res);

						//Add RT_GROUP_ICON
						type_entry.OffsetToData = 0x80000000 | name_level_offset;
						name_entry.OffsetToData = 0x80000000 | lang_level_offset;
						lang_entry.OffsetToData = data_offset;

						name_res.NumberOfIdEntries = 1;
						name_res.NumberOfNamedEntries = 0;
						lang_res.NumberOfIdEntries = 1;
						lang_res.NumberOfNamedEntries = 0;
						memcpy(&rsrc_data[name_level_offset], &name_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
						memcpy(&rsrc_data[lang_level_offset], &lang_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
						name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);
						lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);
						
						//Set name ad ID
						name_entry.Name = 0x1;
						memcpy(&rsrc_data[type_level_offset], &type_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
						memcpy(&rsrc_data[name_level_offset], &name_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
						memcpy(&rsrc_data[lang_level_offset], &lang_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
						memcpy(&rsrc_data[res_offset], image.section_data_from_rva(data.OffsetToData), data.Size);
						
						//Set up icon numbers inside added icon group
						ICON_GROUP* copied_group = reinterpret_cast<ICON_GROUP*>(&rsrc_data[res_offset + sizeof(ICO_HEADER)]);
						for(unsigned long i = 0; i < icon_count; i++)
						{
							copied_group->Number = static_cast<WORD>(i + 1);
							copied_group++;
						}

						data.OffsetToData = rsrc_section->header.VirtualAddress + res_offset;
						memcpy(&rsrc_data[data_offset], &data, sizeof(IMAGE_RESOURCE_DATA_ENTRY));

						type_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
						name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
						lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
						data_offset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);
						res_offset += data.Size;

						name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element
						lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element
					}

					//If image has version info and manifest
					DWORD repacked_resources[] = { reinterpret_cast<DWORD>(RT_VERSION), reinterpret_cast<DWORD>(RT_MANIFEST) };
					BOOST_FOREACH(DWORD type_id, repacked_resources)
					{
						if(get_resource(image, type_id, data, type_entry, name_entry, name_res, lang_entry, lang_res))
						{
							type_entry.OffsetToData = 0x80000000 | name_level_offset;
							name_entry.OffsetToData = 0x80000000 | lang_level_offset;
							lang_entry.OffsetToData = data_offset;

							name_res.NumberOfIdEntries = 1;
							name_res.NumberOfNamedEntries = 0;
							lang_res.NumberOfIdEntries = 1;
							lang_res.NumberOfNamedEntries = 0;
							memcpy(&rsrc_data[name_level_offset], &name_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
							memcpy(&rsrc_data[lang_level_offset], &lang_res, sizeof(IMAGE_RESOURCE_DIRECTORY));
							name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);
							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY);

							//Set name as ID
							name_entry.Name = 0x1;
							memcpy(&rsrc_data[type_level_offset], &type_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
							memcpy(&rsrc_data[name_level_offset], &name_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
							memcpy(&rsrc_data[lang_level_offset], &lang_entry, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY));
							memcpy(&rsrc_data[res_offset], image.section_data_from_rva(data.OffsetToData), data.Size);

							data.OffsetToData = rsrc_section->header.VirtualAddress + res_offset;
							memcpy(&rsrc_data[data_offset], &data, sizeof(IMAGE_RESOURCE_DATA_ENTRY));

							type_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
							name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY);
							data_offset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);
							res_offset += data.Size;

							name_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element
							lang_level_offset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //null element
						}
					}

					//Set up image resource directory address and size to our new resource section
					image.set_directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE, rsrc_section->header.VirtualAddress);
					image.set_directory_size(IMAGE_DIRECTORY_ENTRY_RESOURCE, rsrc_section->get_raw_data().length());
				}
				else
				{
					image.remove_directory(IMAGE_DIRECTORY_ENTRY_RESOURCE);
				}
			}
			else
			{
				//If image has no resources to repack, set up unpacker addresses to 0
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[res_data_dir_addr + size_of_additional_data]) = 0;
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[original_res_addr + size_of_additional_data]) = 0;
				*reinterpret_cast<DWORD*>(&unpacker_raw_data[original_res_size + size_of_additional_data]) = 0;
			}


			//No more data to append, so add section to PE file
			added_section = &image.add_section(unpack_section);



			if(add_import)
			{
				std::cout << "Building fake import directory..." << std::endl;

				std::string& raw_section_data = added_section->get_raw_data();

				old_import_place = import_place;

				IMAGE_IMPORT_DESCRIPTOR d;
				d.OriginalFirstThunk = added_section->header.VirtualAddress + import_place + sizeof(IMAGE_IMPORT_DESCRIPTOR) * 2;
				d.FirstThunk = d.OriginalFirstThunk + sizeof(DWORD) * 2;
				d.Name = d.FirstThunk + sizeof(DWORD) * 2 + sizeof(WORD) + 9; //import by name hint and name
				d.ForwarderChain = 0;
				d.TimeDateStamp = 0;
				memcpy(&raw_section_data[import_place], &d, sizeof(IMAGE_IMPORT_DESCRIPTOR));

				import_place += sizeof(IMAGE_IMPORT_DESCRIPTOR) * 2;

				*reinterpret_cast<DWORD*>(&raw_section_data[import_place]) = d.Name - sizeof(WORD) - 9;
				import_place += sizeof(DWORD) * 2;

				*reinterpret_cast<DWORD*>(&raw_section_data[import_place]) = d.Name + 11;
				import_place += sizeof(DWORD) * 2;
				
				*reinterpret_cast<WORD*>(&raw_section_data[import_place]) = 0x0; //hint
				import_place += sizeof(WORD);
				memcpy(&raw_section_data[import_place], "lstrlenA", 9);
				import_place += 9;
				
				memcpy(&raw_section_data[import_place], "kernel32.dll", 13);
				import_place += 13;
				import_place += sizeof(DWORD);
			}


			//Finish rebuilding exports
			if(image.has_exports())
			{
				std::cout << "Building export table..." << std::endl;

				std::string& raw_section_data = added_section->get_raw_data();

				//Go to start of export table
				char* export_table = &raw_section_data[size_of_additional_data - size_of_exports];
				//Go to our reserved place
				char* exports_place_start = export_table + sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length;

				//Set up library name RVA
				reinterpret_cast<PIMAGE_EXPORT_DIRECTORY>(export_table)->Name = added_section->header.VirtualAddress
					+ size_of_fixups + size_of_tls + sizeof(IMAGE_EXPORT_DIRECTORY);


				//Firstly, rebuild function address table
				unsigned long i = 0;
				unsigned long ordinal = 0;
				for(std::map<WORD, exported_function>::const_iterator it = functions.begin(); it != functions.end(); ++it)
				{
					//Don't forget to add ordinal spacing
					if((*it).first != ordinal)
					{
						for(unsigned long j = ordinal; j < (*it).first; j++)
						{
							*reinterpret_cast<DWORD*>(exports_place_start + i) = 0;
							i += sizeof(DWORD);
							ordinal++;
						}
					}

					*reinterpret_cast<DWORD*>(exports_place_start + i) = (*it).second.rva;
					i += sizeof(DWORD);
					ordinal++;
				}

				//Rebuild name ordinal table
				//Now we have to create name-ordered set of functions
				//because they must be listed alphabetically
				std::set<exported_function> ordered_functions;
				for(std::map<WORD, exported_function>::const_iterator it = functions.begin(); it != functions.end(); ++it)
					ordered_functions.insert((*it).second);

				exports_place_start += ordinal * sizeof(DWORD);
				i = 0;
				for(std::set<exported_function>::const_iterator it = ordered_functions.begin(); it != ordered_functions.end(); ++it)
				{
					if(!(*it).has_name)
						continue;

					*reinterpret_cast<WORD*>(exports_place_start + i) = (*it).name_ordinal;
					i += sizeof(WORD);
				}

				//The most difficult part - rebuild name address table and add names
				exports_place_start += i;
				unsigned long named_function_count = i / sizeof(WORD);
				i = 0;
				unsigned long total_length = 0;
				for(std::set<exported_function>::const_iterator it = ordered_functions.begin(); it != ordered_functions.end(); ++it)
				{
					if(!(*it).has_name)
						continue;

					//Put function name to its place first
					memcpy(exports_place_start + named_function_count * sizeof(DWORD) + total_length, (*it).name.c_str(), (*it).name.length() + 1);

					//Write address of this name to name address table
					*reinterpret_cast<DWORD*>(exports_place_start + i) = added_section->header.VirtualAddress + size_of_fixups + size_of_tls + 
						+ sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length + ordinal * sizeof(DWORD) + named_function_count * (sizeof(DWORD) + sizeof(WORD)) + total_length;

					total_length += (*it).name.length() + 1;

					i += sizeof(DWORD);
				}

				//Now save redirect names and calculate their RVA's
				i = 0;
				for(std::map<WORD, exported_function>::const_iterator it = functions.begin(); it != functions.end(); ++it)
				{
					if(!(*it).second.has_name)
						continue;

					if((*it).second.redirect)
					{
						//Put functio redirect name to its place first
						memcpy(exports_place_start + named_function_count * sizeof(DWORD) + total_length, (*it).second.redirect_name.c_str(), (*it).second.redirect_name.length() + 1);

						//Write address of this name to function address table
						*reinterpret_cast<DWORD*>(export_table + sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length + i) = added_section->header.VirtualAddress + size_of_fixups + size_of_tls + 
							+ sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length + ordinal * sizeof(DWORD) + named_function_count * (sizeof(DWORD) + sizeof(WORD)) + total_length;

						total_length += (*it).second.name.length() + 1;
					}

					i += sizeof(DWORD);
				}

				//Set up address table and name ordinal table addresses
				reinterpret_cast<PIMAGE_EXPORT_DIRECTORY>(export_table)->AddressOfFunctions = added_section->header.VirtualAddress
					+ size_of_fixups + size_of_tls + sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length;
				reinterpret_cast<PIMAGE_EXPORT_DIRECTORY>(export_table)->AddressOfNameOrdinals = added_section->header.VirtualAddress
					+ size_of_fixups + size_of_tls + sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length + ordinal * sizeof(DWORD);

				//Set up name address table address
				reinterpret_cast<PIMAGE_EXPORT_DIRECTORY>(export_table)->AddressOfNames = added_section->header.VirtualAddress
					+ size_of_fixups + size_of_tls + sizeof(IMAGE_EXPORT_DIRECTORY) + library_name_length + ordinal * sizeof(DWORD) + named_function_count * sizeof(WORD);

				//Set up VA and size of section
				image.set_directory_rva(IMAGE_DIRECTORY_ENTRY_EXPORT, added_section->header.VirtualAddress + size_of_fixups + size_of_tls);
				image.set_directory_size(IMAGE_DIRECTORY_ENTRY_EXPORT, size_of_exports);
			}
		}

		std::string& raw_section_data = added_section->get_raw_data();



		//Adjust unpacker section TLS addresses, if PE has TLS
		if(image.has_tls())
		{
			std::cout << "Adjusting TLS..." << std::endl;

			//Map raw TLS data into virtual address space
			PIMAGE_TLS_DIRECTORY tls = reinterpret_cast<PIMAGE_TLS_DIRECTORY>(&added_section->get_virtual_data()[0]);

			//Calculate original TLS data length
			std::size_t len = tls->EndAddressOfRawData - tls->StartAddressOfRawData;

			//Save original address of index in unpacker stub
			*reinterpret_cast<DWORD*>(&raw_section_data[orig_tls_index + size_of_additional_data]) = tls->AddressOfIndex;

			//Fix raw (virtual, actually) data addresses
			tls->StartAddressOfRawData = TLS_ALIGN(image.get_image_base_32() + added_section->header.VirtualAddress + actual_data_length);
			tls->EndAddressOfRawData = tls->StartAddressOfRawData + len;

			//Callbacks are not supported, so null them
			if(tls->AddressOfCallBacks)
			{
				if(image.section_data_length_from_va(tls->AddressOfCallBacks) >= sizeof(DWORD))
				{
					//Warn user if image has TLS callback(s)
					if(*reinterpret_cast<DWORD*>(image.section_data_from_va(tls->AddressOfCallBacks)) != 0)
						std::cout << "WARNING! Image has TLS callback(s), which are not supported!" << std::endl;
				}

				tls->AddressOfCallBacks = 0;
			}

			//Use our empty DWORD, added earlier, to store TLS index
			tls->AddressOfIndex = image.get_image_base_32() + added_section->header.VirtualAddress + sizeof(IMAGE_TLS_DIRECTORY);

			//Save modified index address in unpacker stub
			*reinterpret_cast<DWORD*>(&raw_section_data[tls_index + size_of_additional_data]) = tls->AddressOfIndex;

			//Set up TLS directory RVA
			image.set_directory_rva(IMAGE_DIRECTORY_ENTRY_TLS, added_section->header.VirtualAddress);

			if(image.has_reloc())
			{
				//Don't forget to set relocation directory VA
				PIMAGE_BASE_RELOCATION reloc = reinterpret_cast<PIMAGE_BASE_RELOCATION>(&raw_section_data[size_of_tls]);
				reloc->VirtualAddress = image.directory_rva(IMAGE_DIRECTORY_ENTRY_TLS);
			}
		}
		else
		{
			//Set original & modified TLS indices addresses to NULL
			*reinterpret_cast<DWORD*>(&raw_section_data[orig_tls_index + size_of_additional_data]) = 0;
			*reinterpret_cast<DWORD*>(&raw_section_data[tls_index + size_of_additional_data]) = 0;
		}

		std::cout << std::dec << "Packing data..." << std::endl << std::endl;

		//Pack sections
		unsigned long n = 0; //Current section number
		{
			//Get compression workspace
			unsigned long final, cb1, cb2;
			if(STATUS_SUCCESS != RtlGetCompressionWorkSpaceSize_(COMPRESSION_FORMAT_LZNT1 | COMPRESSION_ENGINE_MAXIMUM, &cb1, &cb2))
				throw pe_exception("Can't get compression workspace");

			boost::scoped_array<char> workbuf(new char[cb1]);

			for(std::vector<pe32::section>::iterator it = image.get_image_sections().begin(); it != image.get_image_sections().end(); ++it)
			{
				//If we reached packer section
				if(&*it == added_section)
					break;

				//Don't pack .rsrc section
				if((*it).get_name() == ".rsrc")
				{
					std::cout << ".rsrc section found, won't be packed..." << std::endl << std::endl;
					continue;
				}

				//Don't pack sections with zero raw sizes
				if(!(*it).get_raw_data().length())
				{
					std::cout << "Zero-sized section found (" << (*it).get_name() << "), won't be packed..." << std::endl << std::endl;
					continue;
				}

				DWORD old_characteristics = (*it).header.Characteristics;
				//Set section to writeable
				(*it).writeable(true);

				if(!pack_sections.empty() && std::find(pack_sections.begin(), pack_sections.end(), (*it).get_name()) == pack_sections.end())
				{
					std::cout << (*it).get_name() << " section won't be packed..." << std::endl << std::endl;
					continue;
				}

				unsigned long original_size = (*it).get_raw_data().length();
				std::cout << "Packing section " << (*it).get_name() << std::endl
					<< "Section size: " << original_size << std::endl;

				//Save section attributes
				*reinterpret_cast<DWORD*>(&raw_section_data[sect_attrs + n * sect_attrs_space + size_of_additional_data]) = old_characteristics;

				//Additional buffer size for compressed data
				unsigned long added_size = 512;

				boost::scoped_array<char> buf;
				while(true)
				{
					//Save section original size
					*reinterpret_cast<DWORD*>(&raw_section_data[sect_orig_sizes + n * sect_orig_sizes_space + size_of_additional_data]) = (*it).get_raw_data().length() + added_size;

					//Compress section data
					buf.reset(new char[(*it).get_raw_data().length() + added_size]);
					DWORD ret = RtlCompressBuffer_(COMPRESSION_FORMAT_LZNT1 | COMPRESSION_ENGINE_MAXIMUM,
						&(*it).get_raw_data()[0],
						(*it).get_raw_data().length(),
						buf.get(),
						(*it).get_raw_data().length() + added_size,
						4096,
						&final,
						workbuf.get());

					if(ret == STATUS_BUFFER_TOO_SMALL)
					{
						std::cout << "Compressed data buffer size too small, increasing..." << std::endl;
						added_size += 4096;
						std::cout <<"Recompressing..." << std::endl;
						continue;
					}
					else if(ret == STATUS_BUFFER_ALL_ZEROS)
					{
						std::cout << "Section contains all zeros, stripping..." << std::endl;
						final = 0;
					}
					else if(ret != STATUS_SUCCESS)
					{
						throw pe_exception("Unable to compress data");
					}

					break;
				}

				if(final)
				{
					//Copy compressed data to section raw data
					(*it).get_raw_data().resize(final, 'A');
					memcpy(&(*it).get_raw_data()[0], buf.get(), final);

					//Save final compressed size
					*reinterpret_cast<DWORD*>(&raw_section_data[sect_sizes + n * sect_size_space + size_of_additional_data]) = final;
				}
				else
				{
					//If section is stripped (all zeros)
					(*it).get_raw_data().clear();
				}

				//Now set up section virtual address
				*reinterpret_cast<DWORD*>(&raw_section_data[sect_va + n * sect_va_space + size_of_additional_data]) = (*it).header.VirtualAddress + image.get_image_base_32();

				std::cout << "Compressed size: " << final << " (" << (100 * static_cast<double>(final) / original_size) << "%)" << std::endl << std::endl;

				if(final)
					n++;
			}
		}

		//Fill left array entries with 0
		for(unsigned int i = n; i < 16; i++)
		{
			*reinterpret_cast<DWORD*>(&raw_section_data[sect_va + i * sect_va_space + size_of_additional_data]) = 0;
			*reinterpret_cast<DWORD*>(&raw_section_data[sect_sizes + i * sect_size_space + size_of_additional_data]) = 0;
			*reinterpret_cast<DWORD*>(&raw_section_data[sect_orig_sizes + i * sect_orig_sizes_space + size_of_additional_data]) = 0;
			*reinterpret_cast<DWORD*>(&raw_section_data[sect_attrs + i * sect_attrs_space + size_of_additional_data]) = 0;
		}

		//Save image base to simplify its search inside unpacker
		*reinterpret_cast<DWORD*>(&raw_section_data[image_base + size_of_additional_data]) = image.get_image_base_32();
		*reinterpret_cast<DWORD*>(&raw_section_data[fixed_image_base + size_of_additional_data]) = image.get_image_base_32();

		//Add fixup to make loader to save HMODULE to our local variable
		if(image.has_reloc())
		{
			DWORD fixup_addr = 
				!image.has_tls() ? 0 :
				size_of_tls + sizeof(IMAGE_BASE_RELOCATION) + sizeof(WORD) * 3;

			fixup_addr += size_of_export_relocs;

			PIMAGE_BASE_RELOCATION reloc = reinterpret_cast<PIMAGE_BASE_RELOCATION>(&raw_section_data[fixup_addr]);
			reloc->VirtualAddress = added_section->header.VirtualAddress + size_of_additional_data + fixed_image_base;
		}

		//Save original entry point
		*reinterpret_cast<DWORD*>(&raw_section_data[OEP + size_of_additional_data]) = image.get_ep() + image.get_image_base_32();

		//Save IAT address
		*reinterpret_cast<DWORD*>(&raw_section_data[IAT + size_of_additional_data]) = 
			image.has_imports() ? image.directory_rva(IMAGE_DIRECTORY_ENTRY_IMPORT) + image.get_image_base_32() : 0;

		//Expand unpacker section virtual size
		//TODO: check virtual size
		image.set_section_virtual_size(*added_section, added_section->header.Misc.VirtualSize + 10000);

		//Remove imports (or add fake import), our unpacker fixes PE import table later
		if(!add_import)
		{
			image.remove_directory(IMAGE_DIRECTORY_ENTRY_IMPORT);
		}
		else
		{
			image.set_directory_rva(IMAGE_DIRECTORY_ENTRY_IMPORT, added_section->header.VirtualAddress + old_import_place);
			image.set_directory_size(IMAGE_DIRECTORY_ENTRY_IMPORT, import_place - old_import_place);
		}

		image.remove_directory(IMAGE_DIRECTORY_ENTRY_IAT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT);

		//TODO. remove config
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);

		//TODO. remove delay imports
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT);

		//If image has relocations, set up Virtual Address of temporary empty fixup table
		if(image.has_reloc())
			image.set_directory_rva(IMAGE_DIRECTORY_ENTRY_BASERELOC, added_section->header.VirtualAddress + size_of_tls);


		//Realign all PE sections
		image.realign_all_sections();

		//Set image entry point to unpacker section
		image.set_ep(image.get_image_sections().back().header.VirtualAddress + size_of_additional_data);


		//Save PE
		std::cout << "Saving image..." << std::endl;
		std::string filename;

		if(output_file_name.empty())
		{
			filename = input_file_name;
			std::string::size_type pos;
			if((pos = filename.find_last_of('.')) != std::string::npos)
				filename = filename.substr(0, pos) + ".packed" + filename.substr(pos);
			else
				filename += "_packed";
		}
		else
		{
			filename = output_file_name;
		}

		std::ofstream out;
		out.open(filename.c_str(), std::ios::out | std::ios::binary);
		image.rebuild_pe(out, strip_dos_headers);


		if(!out)
			std::cout << "Error saving image to file " << filename << std::endl;

		std::cout << "Done in " << tmr.elapsed() << " sec!" << std::endl;
	}
	catch(const pe_exception& e)
	{
		std::cout << std::endl << "Error: " << e.what() << std::endl;
	}

	system("pause");
	return 0;
}
